// @ts-nocheck

const ChildProcessPromise = require("child-process-promise");
const Utils = require("jskippy-node").Utils;
const fs = require("fs-extra");
const RecursiveIterator = require("recursive-iterator");
const bashParser = require("bash-parser");
const sleep = require("sleep-promise");

const path = require("path");
const assert = require("assert");
const os = require("os");
const ChildProcess = require("child_process");

class CommandExecutor
{
	/**
	 * @param {string} strCommand 
	 * @param {Object} [objConfig]
	 * @param {number} [objConfig.nExpectedExitCode = 0]
	 * @param {boolean} [objConfig.bIgnoreParse = true]
	 * @param {boolean} [objConfig.bPipeOutput = true]
	 * @returns {Promise<string>}
	 * @throws {Error}
	 */
	static async execCommandProcess(strCommand, {nExpectedExitCode = 0, bIgnoreParse = true, bPipeOutput = true} = {})
	{
		if(strCommand === "")
		{
			return "";
		}

		const objOptions = {
			capture: [ "stdout", "stderr" ],
			shell: true
		};

		const commandProcessPromise = ChildProcessPromise.spawn(strCommand, [], objOptions);
		const commandProcess = commandProcessPromise.childProcess;

		if(bPipeOutput)
		{
			commandProcess.stdout.pipe(process.stdout);
			commandProcess.stderr.pipe(process.stderr);
		}

		let strCommandOutput = "";

		let objResult = await commandProcessPromise;
		strCommandOutput = objResult.stdout.toString();

		if(objResult.code !== nExpectedExitCode)
		{
			throw new Error(`The expected exit code: ${nExpectedExitCode} does not match with the produced one: ${objResult.code}`);
		}

		// Check result for correct JSON format
		if(!bIgnoreParse)
		{
			Utils.safeJSONParse(strCommandOutput);
		}


		return strCommandOutput;
	}


	/**
	 * Executes a process with params (command line with params) in the context of a shell.
	 * 
	 * It (should/will) is used for ansible and ipmitool commands. 
	 * From a security point of view, both ansible and ipmitool may affect any server with arbitrary commands (reconfigured iLO IPMI & remote desktop and any shell commands on current server), 
	 * including the one on which this code is running. 
	 * 
	 * **Security warning: When using this function, Ansible may affect the machine on which it is executing or at least other Ansible bundles.**
	 * **Security warning: Only run trusted Ansible bundles through this function.**
	 * **Security warning: You can run NOT trusted Ansible bundles through the bsi-ansible-runner service (ansiblerunner, aka AnsibleRunner), which is a separate Datacenter Agent microservice.**
	 * 
	 * There is a executable names whitelist configured. 
	 * Kind of pointless if and when an attacker may inject ansible commands somehow, somewhere.
	 * Otherwise, pretty useful, at least as a deterrent against programmer abuse.
	 * 
	 * On error, the current output recorded from the stdout and stderr streams is included in the error to make debugging of hanging or waiting for input commands much easier.
	 * 
	 * Returns the combined output of stdout and stderr, even when the command crashes or times out.
	 * 
	 * The directory of the files extracted from strTarXZBundleEncrypted is chmod'ed to the 0600 file access mask to secure potential private keys (Ansible requirement at least) and disallow executable scripts.
	 * 
	 * @param {string} strCommand The command to run, with space-separated arguments.
	 * @param {number} nTimeoutSeconds Will send SIGTERM if the process is still running after these many seconds.
	 * @param {string} strTarXZBundleDecrypted = null If specified, it will be extracted to a random /tmp subdirectory and the execution working directory moved inside the extracted directory.
	 * @param {string} strBundleName = "bundle"
	 * @param {Object} objEnvironmentVariables = {}
	 * 
	 * @returns {Promise<{command:string, output:string, exitCode:number, errorMessage:string|null, stdout:string, stderr:string, timedOut:boolean, localCurrentWorkingDirectory:string, currentHostname:string, durationSeconds: number}>}
	 */
	async exec(strCommand, nTimeoutSeconds, strTarXZBundleDecrypted = null, strBundleName = "bundle", objEnvironmentVariables = {})
	{
		assert(typeof strCommand === "string");
		
		let strCurrentWorkingDirectory = os.tmpdir();

		// * @param {number} nMaxBufferBytes = 104857600 Default 100MB. If exceeded the child process is killed.
		const nMaxBufferBytes = 104857600;


		if(strCommand.match(/[\r\n]+/))
		{
			throw new Error("The command contains disallowed new line characters.");
		}


		strBundleName = strBundleName.replace(/[\\\/\.]/g, "_");
		const strAllBundlesDirectoryPath = path.join(os.tmpdir(), "bsi-power-exec-bundles");
		const strBundleDirectoryPath = path.join(strAllBundlesDirectoryPath, `exec_${strBundleName}_${strCommand.split(/[^a-z0-9_\-]/i)[0]}_${new Date().toISOString().replace(/[^0-9Z\.\-]/g, "")}_${Math.random()}`);
		if(strTarXZBundleDecrypted)
		{
			await fs.ensureDir(path.join(os.tmpdir(), "bsi-power-exec-bundles"));

			const strBundleTarXZPath = `${strBundleDirectoryPath}.tar.xz`;
			await fs.writeFile(strBundleTarXZPath, strTarXZBundleDecrypted);
			
			await fs.ensureDir(strBundleDirectoryPath);
			const strUnarchiveCommand = `tar -xvf ${strBundleTarXZPath} -C ${strBundleDirectoryPath}`;
			const objExecutionResult = await this.exec(strUnarchiveCommand, /*nTimeoutSeconds*/ 10);

			// eslint-disable-next-line radix
			if(parseInt(objExecutionResult.exitCode) !== 0 || !fs.existsSync(strBundleDirectoryPath))
			{
				if(!fs.existsSync(strBundleDirectoryPath))
				{
					objExecutionResult.errorMessage = `strBundleDirectoryPath not found: ${strBundleDirectoryPath}`;
					objExecutionResult.exitCode = 1;
				}
				return objExecutionResult;
			}

			await fs.unlink(strBundleTarXZPath);


			const strFileSystemPermissionsCommand = `chmod -R 0600 ${strBundleDirectoryPath}`;
			/*const objExecutionResultFileSystemPermissions =*/ await this.exec(strFileSystemPermissionsCommand, /*nTimeoutSeconds*/ 50);


			strCurrentWorkingDirectory = strBundleDirectoryPath;
		}


		try
		{
			if(os.platform() === "linux")
			{
				const arrWhitelistedExecutableNames = [
					// BSI deploy commands.
					"ansible", "ansible-playbook", "ipmitool", "python", "mkdir", 
					
					// BSI deploy commands, potentially dangerous
					"chmod", "chown", 

					// BSI Kerberos init stuff
					"/usr/local/bin/kadmin", "/usr/local/sbin/kadmin.local", "/usr/bin/k5start", 
					"kadmin", "kadmin.local", "k5start", 

					// Linux read-only tools.
					"grep", "nslookup", "netstat", "traceroute", "whoami", "ping", "md5sum", "df",

					// Needed by this function (exec).
					"tar",

					// Legacy BSI template provisioner iscsiadm stuff
					"parted", "e2fsck", "xfs_repair", "xfs_growfs", "resize2fs", "mount", "umount", 
					"iscsid", "iscsiadm", 
					/*DANGEROUS*/ "rm", 
					/*DANGEROUS*/ "ssh-keygen"

				];
				// Output redirection is considered a command. 
				// bashParser calls it "Redirect". 
				// It is not allowed for security reasons.

				
				// Do not rename "node", it is a property name returned by RecursiveIterator.
				for(let {node/*, path*/} of new RecursiveIterator(bashParser(strCommand)))
				{
					if(typeof node === "object" && node.type === "Command" && typeof node.name === "object")
					{
						if(!arrWhitelistedExecutableNames.includes(node.name.text.toLowerCase()))
						{
							throw new Error(`The sh command ${strCommand} contains an executable name which is not whitelisted: ${node.name.text}. Currently whitelisted executable names: ${JSON.stringify(arrWhitelistedExecutableNames)}`);
						}
					}
				}
			}
			else
			{
				throw new Error(`Currently, the bsi-power exec API is not allowed on ${JSON.stringify(os.platform())} platforms.`);
			}

			const dateStart = new Date();

			const processCommand = ChildProcess.exec(
				strCommand,
				{
					timeout: nTimeoutSeconds * 1000,
					maxBuffer: nMaxBufferBytes,
					cwd: strCurrentWorkingDirectory,
					env: Object.assign({}, objEnvironmentVariables)
				}
			);
			
			processCommand.stdout.setEncoding("utf8");
			processCommand.stderr.setEncoding("utf8");

			const objReturn = {
				currentHostname: os.hostname(),
				command: strCommand,
				environment: objEnvironmentVariables,
				output: "",
				stdout: "",
				stderr: "",
				exitCode: 0,
				errorMessage: null,
				timedOut: false,
				localCurrentWorkingDirectory: strCurrentWorkingDirectory,
				durationSeconds: 0
			};

			processCommand.stdout.on("data", (strData) => {
				objReturn.output += strData;
				objReturn.stdout += strData;
			});
			processCommand.stderr.on("data", (strData) => {
				objReturn.output += strData;
				objReturn.stderr += strData;
			});

			const nStartTimeMilliseconds = new Date().getTime();

			try
			{
				await new Promise(async(fnResolve, fnReject) => {
					processCommand.on("error", fnReject);
					processCommand.on("exit", (nCode) => {
						objReturn.exitCode = nCode;
						objReturn.durationSeconds = parseInt((new Date().getTime() - dateStart.getTime()) / 1000);
						fnResolve();
					});
				});
			}
			catch(error)
			{
				if(objReturn.exitCode === 0)
				{
					objReturn.exitCode = 1;
					// eslint-disable-next-line radix
					objReturn.durationSeconds = parseInt((new Date().getTime() - dateStart.getTime()) / 1000);
					objReturn.errorMessage = error.message;
				}

				if((new Date().getTime() - nStartTimeMilliseconds) / 1000 >= nTimeoutSeconds - /*1 second buffer*/ 1)
				{
					objReturn.timedOut = true;
				}
			}
			finally
			{
				try
				{
					// How could this be?
					if(!processCommand.killed)
					{
						processCommand.kill("SIGTERM");
					}
				}
				catch(error)
				{
					console.error(error);
				}
			}

			// Give time for stdout flush.
			await sleep(2000);
			
			return objReturn;
		}
		finally
		{
			if(fs.existsSync(strBundleDirectoryPath))
			{
				try
				{
					setTimeout(
						async() => {
							try
							{
								await fs.remove(strBundleDirectoryPath);
							}
							catch(error)
							{
								console.error(error);
							}
						}, 
						(1 * 3600) * 1000
					);
				}
				catch(error)
				{
					console.error(error);
				}
			}
		}
	}
}

module.exports = CommandExecutor;
