//@ts-check

const CommandExecutor = require("../CommandExecutor");

class IPMISetup
{
	constructor()
	{}

	/**
	 * This function will set up IPMI for the server.
	 * IPMI credentials can be pregenerated or allocated by the agent.
	 * 
	 * @param {Object} [objConfig]
	 * @param {Encryption} serverEncryption
	 * @param {boolean} bOnlyEnableIPMIOverLan
	 * @returns {Promise<Object>}
	 */
	static async run(objConfig, serverEncryption, bOnlyEnableIPMIOverLan)
	{
		const strIPMICfgLocation = "IPMICFG=/usr/bin/ipmicfg";
		const strIPMIToolLocation = "IPMITOOL=/usr/bin/ipmitool";

		let strCommand = strIPMIToolLocation + `
			for channel in 0 1 2 3 4
			do
				result=\`$IPMITOOL lan print $channel\`
				if [ -n "$result" ]; then
					ipmichannel=$channel
				fi
			done
			echo "${IPMISetup.START_IPMICHANNEL}\${ipmichannel}${IPMISetup.END_IPMICHANNEL}"
		`;

		const nIPMIChannel = await IPMISetup._executeCommandAndParse(strCommand, {bResetBMC: false, bParseIPMIChannel: true});

		if(bOnlyEnableIPMIOverLan)
		{
			strCommand = strIPMIToolLocation + `
				$IPMITOOL lan set ${nIPMIChannel} access on
			`;

			await IPMISetup._executeCommandAndParse(strCommand, {bResetBMC: false, bParseIPMIChannel: false});
			return {};
		}

		const {
			objServer,
			objOOBSubnet,
			strIPMIUsername,
			strIPMIPassword,
			strInternalAdministratorUserID = IPMISetup.INTERNAL_ADMINISTRATOR_USER_ID
		} = objConfig;


		if(
			objServer 
			&& objServer.server_status !== IPMISetup.SERVER_STATUS_USED_REGISTERING
		)
		{
			if(!objOOBSubnet)
			{
				throw new Error("Empty object objOOBSubnet. Must provide OOBSubnet information.");
			}

			await IPMISetup._executeCommandAndParse(
				strIPMIToolLocation + `
					$IPMITOOL user set name ${strInternalAdministratorUserID} ${strIPMIUsername}
				`, 
				{
					bResetBMC: false, 
					bParseIPMIChannel: false
				}
			);

			await IPMISetup._executeCommandAndParse(
				strIPMIToolLocation + `
					$IPMITOOL user set password ${strInternalAdministratorUserID} ${strIPMIPassword}
				`, 
				{
					bResetBMC: false, 
					bParseIPMIChannel: false
				}
			);


			// Special case for supermicro vendor servers
			if(objServer.server_vendor === "Supermicro")
			{
				await IPMISetup._executeCommandAndParse(
					strIPMICfgLocation + `
						$IPMICFG -user setpwd ${strInternalAdministratorUserID} ${strIPMIPassword}
					`, 
					{
						bResetBMC: false, 
						bParseIPMIChannel: false
					}
				);
			}

			await IPMISetup._executeCommandAndParse(
				strIPMIToolLocation + `
					$IPMITOOL user priv ${strInternalAdministratorUserID} ${IPMISetup.PRIVILEGE_ADMINISTRATOR} ${nIPMIChannel}
				`, 
				{
					bResetBMC: false, 
					bParseIPMIChannel: false
				}
			);

			await IPMISetup._executeCommandAndParse(
				strIPMIToolLocation + `
					$IPMITOOL lan set ${nIPMIChannel} ipsrc static
				`, 
				{
					bResetBMC: false, 
					bParseIPMIChannel: false
				}
			);

			await IPMISetup._executeCommandAndParse(
				strIPMIToolLocation + `
					$IPMITOOL lan set ${nIPMIChannel} ipaddr ${objServer["server_ipmi_host"]}
				`, 
				{
					bResetBMC: false, 
					bParseIPMIChannel: false
				}
			);

			await IPMISetup._executeCommandAndParse(
				strIPMIToolLocation + `
					$IPMITOOL lan set ${nIPMIChannel} netmask ${objOOBSubnet["subnet_oob_netmask_human_readable"]}
				`, 
				{
					bResetBMC: false, 
					bParseIPMIChannel: false
				}
			);

			await IPMISetup._executeCommandAndParse(
				strIPMIToolLocation + `
					$IPMITOOL lan set ${nIPMIChannel} defgw ipaddr ${objOOBSubnet["subnet_oob_gateway_human_readable"]}
				`, 
				{
					bResetBMC: false, 
					bParseIPMIChannel: false
				}
			);

			try
			{
				await IPMISetup._executeCommandAndParse(
					strIPMIToolLocation + `
						$IPMITOOL lan set ${nIPMIChannel} access on
					`, 
					{
						bResetBMC: false, 
						bParseIPMIChannel: false
					}
				);
			}
			catch(error)
			{
				console.error(error.stack);
			}

			// Disabling default administrator user
			// Not all servers have this user
			try
			{
				await IPMISetup._executeCommandAndParse(
					strIPMIToolLocation + `
						$IPMITOOL user disable ${IPMISetup.DEFAULT_ADMINISTRATOR_USER_ID}
					`, 
					{
						bResetBMC: false, 
						bParseIPMIChannel: false
					}
				);
			}
			catch(error)
			{
				console.error(error.stack);
			}

			// Disabling the extra client user
			try
			{
				await IPMISetup._executeCommandAndParse(
					strIPMIToolLocation + `
						$IPMITOOL user disable ${IPMISetup.DEFAULT_CLIENT_USER_ID}
					`, 
					{
						bResetBMC: false, 
						bParseIPMIChannel: false
					}
				);
			}
			catch(error)
			{
				console.error(error.stack);
			}

			await IPMISetup._executeCommandAndParse(
				strIPMIToolLocation + `
					$IPMITOOL user enable ${strInternalAdministratorUserID}
				`, 
				{
					bResetBMC: true, 
					bParseIPMIChannel: false
				}
			);
		}


		const objIPMISettings = {
			server_ipmi_internal_username: strIPMIUsername,
			server_ipmi_internal_password_decrypted: strIPMIPassword,
			server_ipmi_version: 2,
			server_ipmi_channel: nIPMIChannel,
			datacenter_name: objServer.datacenter_name,
			command: strCommand
		};

		return objIPMISettings;
	}


	static allocateIPMIUsername()
	{
		return `bsi${IPMISetup.allocateIPMIPassword(5)}`;
	}

	/**
	* Generates a 16 characters password for IPMI server authentication.
	* Should be used by IPMI setup functions.
	*
	* @param {number} nPasswordLength
	* @returns {string}
	*/
	static allocateIPMIPassword(nPasswordLength = 16)
	{
		return IPMISetup._generateRandomString(nPasswordLength);
	}


	/**
	 * Generates a random string.
	 * characterList - possible characters
	 *
	 * @param {number} [nLength=16]
	 * @param {{characterList: string}|any} [objOptions={}]
	 *
	 * @returns {string}
	 */
	static _generateRandomString(nLength = 16, objOptions = {})
	{
		const {
			characterList = "23456789abcdefghijkmnpqrstuvwxyzABCDEFGHJKLMNPQRSTUVWXYZ" //Not present characters: o, O, 0 (zero), 1 (one), l (L).
		} = objOptions;


		let strRandomText = "";
		for(let i = 0; i < nLength; i++)
		{
			strRandomText += characterList.charAt(Math.floor(Math.random() * characterList.length));
		}

		return strRandomText;
	}


	/**
	 * Executes the given commands and parses the output to extract the ipmichannel.
	 * 
	 * @param {string} strCommand 
	 * @param {Object} objConfig 
	 * @returns {Promise<number>}
	 */
	static async _executeCommandAndParse(strCommand, {bResetBMC = false, bParseIPMIChannel = true} = {})
	{
		console.log(strCommand);
		const strCommandOutput = await CommandExecutor.execCommandProcess(strCommand);

		if(bResetBMC)
		{
			try
			{
				try
				{
					// Executing this command could throw, because some servers fail to execute a warm reset
					await IPMISetup._BMCWarmReset();
				}
				catch(error)
				{
					await IPMISetup._BMCColdReset();
				}
			}
			catch(error)
			{
				console.error(error.stack);
			}
		}

		if(bParseIPMIChannel)
		{
			const strIPMIChannel = strCommandOutput
				.split(IPMISetup.START_IPMICHANNEL)
				.pop()
				.split(IPMISetup.END_IPMICHANNEL)
				.shift();
			
			const nIPMIChannel = parseInt(strIPMIChannel, 10);

			if(Number.isNaN(nIPMIChannel))
			{
				throw new Error(`
					IPMI channel received from the setup script is invalid: ${strCommandOutput}
					Command: ${strCommand}
				`);
			}

			return nIPMIChannel;
		}
	}

	/**
	 * Reset the management controller.
	 * 
	 * @returns {Promise<>}
	 * @throws Error
	 */
	static async _BMCWarmReset()
	{
		const strCommand = `
			IPMITOOL=/usr/bin/ipmitool
			$IPMITOOL bmc reset warm
		`;
		
		await CommandExecutor.execCommandProcess(strCommand);
	}


	/**
	 * Reset the management controller.
	 * 
	 * @returns {Promise<>}
	 * @throws Error
	 */
	static async _BMCColdReset()
	{
		const strCommand = `
			IPMITOOL=/usr/bin/ipmitool
			$IPMITOOL bmc reset cold
		`;
		
		await CommandExecutor.execCommandProcess(strCommand);
	}

	/**
	 * @returns {string}
	 */
	static get SERVER_STATUS_USED_REGISTERING()
	{
		return "used_registering";
	}

	/**
	 * @returns {number}
	 */
	static get PRIVILEGE_ADMINISTRATOR()
	{
		return 4;
	}

	/**
	 * Function that returns the id of the client user in the user list. This user is provisioned when the server is given to a client.
	 * The default value for this is 3.
	 *
	 * @returns {number}
	*/
	static get DEFAULT_CLIENT_USER_ID()
	{
		return 3;
	}

	/**
	 * Function that returns the id of the internal Administrator user in the user list. This user is always present to be used to control the server.
	 * The default value for this is 2.
	 *
	 * @returns {number}
	 */
	static get INTERNAL_ADMINISTRATOR_USER_ID()
	{
		return 2;
	}

	/**
	 * Function that returns the id of the default Administrator user in the user list. This user is present after a factory reset of the server.
	 * The default value for this is 1.
	 *
	 * @returns {number}}
	*/
	static get DEFAULT_ADMINISTRATOR_USER_ID()
	{
		return 1;
	}

	/**
	 * @returns {string}
	 */
	static get START_IPMICHANNEL()
	{
		return "START_IPMICHANNEL";
	}
	
	/**
	 * @returns {string}
	 */
	static get END_IPMICHANNEL()
	{
		return "END_IPMICHANNEL";
	}
}

module.exports = IPMISetup;
